@@ -6,7 +6,7 @@ use std::io::{Error, ErrorKind, Read, Result, Seek, SeekFrom};
 use std::process::{Child, ChildStdout, Command, Stdio};
 
 #[derive(Debug, PartialEq)]
-pub enum Compression {
+pub(crate) enum Compression {
     Uncompressed,
     Bzip2 {
         level: Option<u32>,
@@ -36,7 +36,7 @@ pub enum Compression {
 }
 
 impl Compression {
-    pub fn from_magic_number(magic_number: [u8; 4]) -> Result<Self> {
+    pub(crate) fn from_magic_number(magic_number: [u8; 4]) -> Result<Self> {
         let compression = match magic_number {
             [0x42, 0x5A, 0x68, _] => Compression::Bzip2 { level: None },
             [0x30, 0x37, 0x30, 0x37] => Compression::Uncompressed,
@@ -104,7 +104,7 @@ impl Compression {
         };
     }
 
-    pub fn from_command_line(line: &str) -> Result<Self> {
+    pub(crate) fn from_command_line(line: &str) -> Result<Self> {
         let mut iter = line.split_whitespace();
         let mut compression = if let Some(cmd) = iter.next() {
             Self::from_str(cmd)?
@@ -155,7 +155,7 @@ impl Compression {
         Ok(compression)
     }
 
-    pub fn command(&self) -> &str {
+    pub(crate) fn command(&self) -> &str {
         match self {
             Self::Uncompressed => "cpio",
             Self::Bzip2 { level: _ } => "bzip2",
@@ -172,8 +172,13 @@ impl Compression {
         }
     }
 
-    pub fn compress(&self, file: Option<File>, source_date_epoch: Option<u32>) -> Result<Child> {
-        let mut command = self.compress_command(source_date_epoch);
+    pub(crate) fn compress(
+        &self,
+        file: Option<File>,
+        source_date_epoch: Option<u32>,
+        size: impl FnOnce() -> u64,
+    ) -> Result<Child> {
+        let mut command = self.compress_command(source_date_epoch, size);
         // TODO: Propper error message if spawn fails
         command.stdin(Stdio::piped());
         if let Some(file) = file {
@@ -189,7 +194,11 @@ impl Compression {
         Ok(cmd)
     }
 
-    fn compress_command(&self, source_date_epoch: Option<u32>) -> Command {
+    fn compress_command(
+        &self,
+        source_date_epoch: Option<u32>,
+        size: impl FnOnce() -> u64,
+    ) -> Command {
         let mut command = Command::new(self.command());
         match self {
             Self::Gzip { level: _ } => {
@@ -246,10 +255,13 @@ impl Compression {
         {
             command.arg("-T1");
         }
+        if matches!(self, Self::Zstd { level: _ }) {
+            command.arg(format!("--stream-size={}", size()));
+        }
         command
     }
 
-    pub fn decompress(&self, file: File) -> Result<ChildStdout> {
+    pub(crate) fn decompress(&self, file: File) -> Result<ChildStdout> {
         let mut command = self.decompress_command();
         // TODO: Propper error message if spawn fails
         let cmd = command
@@ -288,35 +300,25 @@ impl Compression {
         command
     }
 
-    pub fn is_uncompressed(&self) -> bool {
+    pub(crate) fn is_uncompressed(&self) -> bool {
         matches!(self, Self::Uncompressed)
     }
 }
 
-pub fn read_magic_header<R: Read + Seek>(file: &mut R) -> Option<Result<Compression>> {
+pub(crate) fn read_magic_header<R: Read + Seek>(file: &mut R) -> Result<Option<Compression>> {
     let mut buffer = [0; 4];
     while buffer == [0, 0, 0, 0] {
         match file.read_exact(&mut buffer) {
             Ok(()) => {}
             Err(e) => match e.kind() {
-                ErrorKind::UnexpectedEof => return None,
-                _ => return Some(Err(e)),
+                ErrorKind::UnexpectedEof => return Ok(None),
+                _ => return Err(e),
             },
         };
     }
-    match file.seek(SeekFrom::Current(-4)) {
-        Ok(_) => {}
-        Err(e) => {
-            return Some(Err(e));
-        }
-    };
-    let compression = match Compression::from_magic_number(buffer) {
-        Ok(compression) => compression,
-        Err(e) => {
-            return Some(Err(e));
-        }
-    };
-    Some(Ok(compression))
+    file.seek(SeekFrom::Current(-4))?;
+    let compression = Compression::from_magic_number(buffer)?;
+    Ok(Some(compression))
 }
 
 #[cfg(test)]
