@@ -188,6 +188,62 @@ fn from_mtime(mtime: u32) -> SystemTime {
     std::time::UNIX_EPOCH + std::time::Duration::from_secs(mtime.into())
 }
 
+fn extract_to_disk<R: Read + SeekForward, W: Write>(
+    archive: &mut R,
+    header: &Header,
+    extractor: &mut Extractor,
+    options: &ExtractOptions,
+    logger: &mut Logger<W>,
+) -> Result<()> {
+    match header.mode & MODE_FILETYPE_MASK {
+        FILETYPE_BLOCK_DEVICE | FILETYPE_CHARACTER_DEVICE | FILETYPE_FIFO | FILETYPE_SOCKET => {
+            write_special_file(header, options.preserve_permissions, logger)?
+        }
+        FILETYPE_DIRECTORY => write_directory(
+            header,
+            options.preserve_permissions,
+            logger,
+            &mut extractor.mtimes,
+        )?,
+        FILETYPE_REGULAR_FILE => write_file(
+            archive,
+            header,
+            options.preserve_permissions,
+            &mut extractor.seen_files,
+            logger,
+        )?,
+        FILETYPE_SYMLINK => {
+            write_symbolic_link(archive, header, options.preserve_permissions, logger)?
+        }
+        _ => {
+            return Err(Error::new(
+                ErrorKind::InvalidData,
+                format!(
+                    "Invalid/unknown file type 0o{:o} for '{}'",
+                    header.mode, header.filename
+                ),
+            ))
+        }
+    }
+    Ok(())
+}
+
+fn extract_to_writable<R, W>(archive: &mut R, header: &Header, out: &mut W) -> Result<()>
+where
+    R: Read + SeekForward,
+    W: Write,
+{
+    if header.filesize == 0 {
+        return Ok(());
+    }
+    if matches!(header.mode & MODE_FILETYPE_MASK, FILETYPE_REGULAR_FILE) {
+        write_file_content(archive, out, header)?;
+    } else {
+        header.skip_file_content(archive)?;
+    }
+    Ok(())
+}
+
 fn read_cpio_and_extract<R: Read + SeekForward, W: Write, LW: Write>(
     archive: &mut R,
     base_dir: &PathBuf,
@@ -234,77 +290,16 @@ fn read_cpio_and_extract<R: Read + SeekForward, W: Write, LW: Write>(
         }
 
         if let Some(out) = out {
-            if header.filesize == 0 {
-                continue;
-            }
-            match header.mode & MODE_FILETYPE_MASK {
-                FILETYPE_DIRECTORY | FILETYPE_SYMLINK => {
-                    header.skip_file_content(archive)?;
-                }
-                FILETYPE_REGULAR_FILE => write_file_content(archive, out, &header)?,
-                FILETYPE_CHARACTER_DEVICE
-                | FILETYPE_FIFO
-                | FILETYPE_BLOCK_DEVICE
-                | FILETYPE_SOCKET => {
-                    unimplemented!(
-                        "Mode {:o} (file {}) not implemented. Please open a bug report requesting support for this type.",
-                        header.mode, header.filename
-                    )
-                }
-                _ => {
-                    return Err(Error::new(
-                        ErrorKind::InvalidData,
-                        format!(
-                            "Invalid/unknown filetype {:o}: {}",
-                            header.mode, header.filename
-                        ),
-                    ))
-                }
-            }
+            extract_to_writable(archive, &header, out)?;
         } else {
-            match header.mode & MODE_FILETYPE_MASK {
-                FILETYPE_CHARACTER_DEVICE => {
-                    write_character_device(&header, options.preserve_permissions, logger)?
-                }
-                FILETYPE_DIRECTORY => write_directory(
-                    &header,
-                    options.preserve_permissions,
-                    logger,
-                    &mut extractor.mtimes,
-                )?,
-                FILETYPE_REGULAR_FILE => write_file(
-                    archive,
-                    &header,
-                    options.preserve_permissions,
-                    &mut extractor.seen_files,
-                    logger,
-                )?,
-                FILETYPE_SYMLINK => {
-                    write_symbolic_link(archive, &header, options.preserve_permissions, logger)?
-                }
-                FILETYPE_FIFO | FILETYPE_BLOCK_DEVICE | FILETYPE_SOCKET => {
-                    unimplemented!(
-                        "Mode {:o} (file {}) not implemented. Please open a bug report requesting support for this type.",
-                        header.mode, header.filename
-                    )
-                }
-                _ => {
-                    return Err(Error::new(
-                        ErrorKind::InvalidData,
-                        format!(
-                            "Invalid/unknown filetype {:o}: {}",
-                            header.mode, header.filename
-                        ),
-                    ))
-                }
-            }
+            extract_to_disk(archive, &header, &mut extractor, options, logger)?;
         };
     }
     extractor.set_modified_times(logger)?;
     Ok(())
 }
 
-fn write_character_device<W: Write>(
+fn write_special_file<W: Write>(
     header: &Header,
     preserve_permissions: bool,
     logger: &mut Logger<W>,
@@ -313,14 +308,17 @@ fn write_character_device<W: Write>(
         return Err(Error::new(
             ErrorKind::InvalidData,
             format!(
-                "Invalid size for character device '{}': {} bytes instead of 0.",
-                header.filename, header.filesize
+                "Invalid size for {} '{}': {} bytes instead of 0.",
+                header.file_type_name(),
+                header.filename,
+                header.filesize
             ),
         ));
     };
     debug!(
         logger,
-        "Creating character device '{}' with mode {:o}",
+        "Creating {} '{}' with mode {:o}",
+        header.file_type_name(),
         header.filename,
         header.mode_perm(),
     )?;
@@ -662,6 +660,76 @@ mod tests {
         assert_eq!(logger.get_logs(), ".\npath\npath/file\n");
     }
 
+    #[test]
+    fn test_read_cpio_and_extract_fifo() {
+        let _lock = TEST_LOCK.lock().unwrap();
+        let tempdir = TempDir::new_and_set_current_dir().unwrap();
+        let path = tempdir.path.join("fifo.cpio");
+        let uid = getuid();
+        let gid = getgid();
+        let header = Header::new(1, 0o010_600, uid, gid, 1, 1746789067, 0, 0, 0, "initctl");
+        let mut archive = File::create(&path).unwrap();
+        header.write(&mut archive).unwrap();
+        Header::trailer().write(&mut archive).unwrap();
+
+        let mut archive = File::open(&path).unwrap();
+        let mut logger = Logger::new_vec(Level::Info);
+        read_cpio_and_extract(
+            &mut archive,
+            &tempdir.path,
+            &mut None::<Stdout>,
+            &ExtractOptions::default(),
+            &mut logger,
+        )
+        .unwrap();
+
+        let attr = std::fs::metadata("initctl").unwrap();
+        assert_eq!(attr.len(), header.filesize.into());
+        assert!(attr.file_type().is_fifo());
+        assert_eq!(attr.modified().unwrap(), from_mtime(header.mtime));
+        assert_eq!(attr.permissions(), PermissionsExt::from_mode(header.mode));
+        assert_eq!(attr.uid(), header.uid);
+        assert_eq!(attr.gid(), header.gid);
+        assert_eq!(major(attr.rdev()), header.rmajor);
+        assert_eq!(minor(attr.rdev()), header.rminor);
+        assert_eq!(logger.get_logs(), "initctl\n");
+    }
+
+    #[test]
+    fn test_read_cpio_and_extract_invalid_file_type() {
+        let tempdir = TempDir::new().unwrap();
+        let path = tempdir.path.join("invalid.cpio");
+        let mut archive = File::create(&path).unwrap();
+        archive
+            .write_all(
+                b"070701000000010003FFA200000007000000070000000168AEBD2C\
+                00000000000000000000000000000000000000000000000800000000\
+                invalid\0\0\0\
+                070701000000000000000000000000000000000000000100000000\
+                00000000000000000000000000000000000000000000000B00000000\
+                TRAILER!!!\0\0\0\0",
+            )
+            .unwrap();
+        let mut archive = File::open(&path).unwrap();
+        let mut logger = Logger::new_vec(Level::Warning);
+        let got = read_cpio_and_extract(
+            &mut archive,
+            &tempdir.path,
+            &mut None::<Stdout>,
+            &ExtractOptions::default(),
+            &mut logger,
+        )
+        .unwrap_err();
+
+        assert_eq!(got.kind(), ErrorKind::InvalidData);
+        assert_eq!(
+            got.to_string(),
+            "Invalid/unknown file type 0o777642 for 'invalid'",
+        );
+        assert_eq!(logger.get_logs(), "");
+        assert!(!Path::new("invalid").exists());
+    }
+
     // Test detecting path traversal attacks like CVE-2015-1197
     #[test]
     fn test_read_cpio_and_extract_path_traversal() {
@@ -704,18 +772,43 @@ mod tests {
     }
 
     #[test]
-    fn test_write_character_device() {
-        let _lock = TEST_LOCK.lock().unwrap();
+    fn test_write_special_file_block_device() {
         if getuid() != 0 {
             // This test needs to run as root.
             return;
         }
+        let _lock = TEST_LOCK.lock().unwrap();
         let _tempdir = TempDir::new_and_set_current_dir().unwrap();
-        let mut header = Header::new(1, 0o20_644, 0, 0, 0, 1740402179, 0, 0, 0, "./null");
-        header.rmajor = 1;
-        header.rminor = 3;
+        let header = Header::new(1, 0o60_660, 0, 6, 1, 1751300235, 0, 7, 99, "loop99");
         let mut logger = Logger::new_vec(Level::Debug);
-        write_character_device(&header, true, &mut logger).unwrap();
+        write_special_file(&header, true, &mut logger).unwrap();
+
+        let attr = std::fs::metadata("loop99").unwrap();
+        assert_eq!(attr.len(), header.filesize.into());
+        assert!(attr.file_type().is_block_device());
+        assert_eq!(attr.modified().unwrap(), from_mtime(header.mtime));
+        assert_eq!(attr.permissions(), PermissionsExt::from_mode(header.mode));
+        assert_eq!(attr.uid(), header.uid);
+        assert_eq!(attr.gid(), header.gid);
+        assert_eq!(major(attr.rdev()), header.rmajor);
+        assert_eq!(minor(attr.rdev()), header.rminor);
+        assert_eq!(
+            logger.get_logs(),
+            "Creating block device 'loop99' with mode 660\n"
+        );
+    }
+
+    #[test]
+    fn test_write_special_file_character_device() {
+        if getuid() != 0 {
+            // This test needs to run as root.
+            return;
+        }
+        let _lock = TEST_LOCK.lock().unwrap();
+        let _tempdir = TempDir::new_and_set_current_dir().unwrap();
+        let header = Header::new(1, 0o20_644, 0, 0, 0, 1740402179, 0, 1, 3, "./null");
+        let mut logger = Logger::new_vec(Level::Debug);
+        write_special_file(&header, true, &mut logger).unwrap();
 
         let attr = std::fs::metadata("null").unwrap();
         assert_eq!(attr.len(), header.filesize.into());
@@ -733,6 +826,53 @@ mod tests {
         std::fs::remove_file("null").unwrap();
     }
 
+    #[test]
+    fn test_write_special_file_fifo() {
+        let _lock = TEST_LOCK.lock().unwrap();
+        let _tempdir = TempDir::new_and_set_current_dir().unwrap();
+        let uid = getuid();
+        let gid = getgid();
+        let header = Header::new(1, 0o010_600, uid, gid, 1, 1746789067, 0, 0, 0, "initctl");
+        let mut logger = Logger::new_vec(Level::Debug);
+        write_special_file(&header, false, &mut logger).unwrap();
+
+        let attr = std::fs::metadata("initctl").unwrap();
+        assert_eq!(attr.len(), header.filesize.into());
+        assert!(attr.file_type().is_fifo());
+        assert_eq!(attr.modified().unwrap(), from_mtime(header.mtime));
+        assert_eq!(attr.permissions(), PermissionsExt::from_mode(header.mode));
+        assert_eq!(attr.uid(), header.uid);
+        assert_eq!(attr.gid(), header.gid);
+        assert_eq!(major(attr.rdev()), header.rmajor);
+        assert_eq!(minor(attr.rdev()), header.rminor);
+        assert_eq!(logger.get_logs(), "Creating fifo 'initctl' with mode 600\n");
+    }
+
+    #[test]
+    fn test_write_special_file_socket() {
+        let _lock = TEST_LOCK.lock().unwrap();
+        let _tempdir = TempDir::new_and_set_current_dir().unwrap();
+        let uid = getuid();
+        let gid = getgid();
+        let header = Header::new(1, 0o140_777, uid, gid, 1, 1746789058, 0, 0, 0, "notify");
+        let mut logger = Logger::new_vec(Level::Debug);
+        write_special_file(&header, true, &mut logger).unwrap();
+
+        let attr = std::fs::metadata("notify").unwrap();
+        assert_eq!(attr.len(), header.filesize.into());
+        assert!(attr.file_type().is_socket());
+        assert_eq!(attr.modified().unwrap(), from_mtime(header.mtime));
+        assert_eq!(attr.permissions(), PermissionsExt::from_mode(header.mode));
+        assert_eq!(attr.uid(), header.uid);
+        assert_eq!(attr.gid(), header.gid);
+        assert_eq!(major(attr.rdev()), header.rmajor);
+        assert_eq!(minor(attr.rdev()), header.rminor);
+        assert_eq!(
+            logger.get_logs(),
+            "Creating socket 'notify' with mode 777\n"
+        );
+    }
+
     #[test]
     fn test_write_directory_with_setuid() {
         let _lock = TEST_LOCK.lock().unwrap();
@@ -835,8 +975,9 @@ mod tests {
             0,
             "./dead_symlink",
         );
+        assert_eq!(header.file_type_name(), "symlink");
         let cpio = b"/nonexistent";
-        let mut logger = Logger::new_vec(Level::Warning);
+        let mut logger = Logger::new_vec(Level::Debug);
         write_symbolic_link(&mut cpio.as_ref(), &header, true, &mut logger).unwrap();
 
         let attr = std::fs::symlink_metadata("dead_symlink").unwrap();
@@ -846,7 +987,10 @@ mod tests {
         assert_eq!(attr.permissions(), PermissionsExt::from_mode(header.mode));
         assert_eq!(attr.uid(), header.uid);
         assert_eq!(attr.gid(), header.gid);
-        assert_eq!(logger.get_logs(), "");
+        assert_eq!(
+            logger.get_logs(),
+            "Creating symlink './dead_symlink' -> '/nonexistent' with mode 777\n"
+        );
         std::fs::remove_file("dead_symlink").unwrap();
     }
 }
