@@ -88,6 +88,19 @@ impl Header {
         (u64::from(self.major) << 32) | u64::from(self.minor)
     }
 
+    pub(crate) fn file_type_name(&self) -> &str {
+        match self.mode & MODE_FILETYPE_MASK {
+            FILETYPE_FIFO => "fifo",
+            FILETYPE_CHARACTER_DEVICE => "character device",
+            FILETYPE_DIRECTORY => "directory",
+            FILETYPE_BLOCK_DEVICE => "block device",
+            FILETYPE_REGULAR_FILE => "regular file",
+            FILETYPE_SYMLINK => "symlink",
+            FILETYPE_SOCKET => "socket",
+            _ => "invalid/unknown file type",
+        }
+    }
+
     pub(crate) fn is_root_directory(&self) -> bool {
         self.filename == "." && self.mode & MODE_FILETYPE_MASK == FILETYPE_DIRECTORY
     }
@@ -390,6 +403,7 @@ mod tests {
                 filename: "path/file".into()
             }
         );
+        assert_eq!(header.file_type_name(), "regular file");
 
         // Test writing the header and get the original data back
         let mut output = Vec::new();
@@ -403,6 +417,37 @@ mod tests {
         assert_eq!(size, archive.len() as u64);
     }
 
+    #[test]
+    fn test_header_read_invalid_file_type() {
+        // Wrapped before mtime and filename
+        let archive = b"070701000000010003FFA200000007000000420000000168AEBD2C\
+            00000000000000000000000000000000000000000000000800000000\
+            invalid\0\0\0";
+        let header = Header::read(&mut archive.as_ref()).unwrap();
+        assert_eq!(
+            header,
+            Header {
+                ino: 1,
+                mode: 0o777642,
+                uid: 7,
+                gid: 0x42,
+                nlink: 1,
+                mtime: 0x68AEBD2C,
+                filesize: 0,
+                major: 0,
+                minor: 0,
+                rmajor: 0,
+                rminor: 0,
+                filename: "invalid".into()
+            }
+        );
+        assert_eq!(header.file_type_name(), "invalid/unknown file type");
+        assert_eq!(
+            std::str::from_utf8(&header.mode_string()).unwrap(),
+            "?rwSr-S-wT"
+        );
+    }
+
     #[test]
     fn test_header_read_invalid_magic_number() {
         let invalid_data = b"abc\tefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ\
@@ -519,6 +564,7 @@ mod tests {
     fn test_is_root_directory() {
         let header = Header::new(0, 0o040_755, 0, 0, 1, 1744150584, 0, 0, 0, ".");
         assert!(header.is_root_directory());
+        assert_eq!(header.file_type_name(), "directory");
     }
 
     #[test]
