@@ -4,19 +4,27 @@
 use std::collections::HashMap;
 use std::fs::File;
 use std::io::{prelude::*, Result};
+use std::num::NonZeroU32;
 use std::time::SystemTime;
 
 use glob::Pattern;
 
 use crate::compression::read_magic_header;
 use crate::filetype::*;
-use crate::header::{read_filename_from_next_cpio_object, Header};
+use crate::header::{
+    read_file_name_and_size_from_next_cpio_object, Header, CPIO_ALIGNMENT, TRAILER_FILENAME,
+};
 use crate::libc::strftime_local;
+use crate::logger::{Level, Logger};
 use crate::manifest::Manifest;
 use crate::ranges::Ranges;
 use crate::seek_forward::SeekForward;
 
+#[macro_use]
+pub mod logger;
+
 mod compression;
+pub mod examine;
 mod extended_error;
 pub mod extract;
 mod filetype;
@@ -27,10 +35,6 @@ pub mod ranges;
 mod seek_forward;
 pub mod temp_dir;
 
-pub const LOG_LEVEL_WARNING: u32 = 5;
-pub const LOG_LEVEL_INFO: u32 = 7;
-pub const LOG_LEVEL_DEBUG: u32 = 8;
-
 struct CpioFilenameReader<'a, R: Read + SeekForward> {
     archive: &'a mut R,
 }
@@ -39,15 +43,15 @@ impl<R: Read + SeekForward> Iterator for CpioFilenameReader<'_, R> {
     type Item = Result<String>;
 
     fn next(&mut self) -> Option<Self::Item> {
-        match read_filename_from_next_cpio_object(self.archive) {
-            Ok(filename) => {
-                if filename == "TRAILER!!!" {
+        match read_file_name_and_size_from_next_cpio_object(self.archive) {
+            Ok((filename, _)) => {
+                if filename == TRAILER_FILENAME {
                     None
                 } else {
                     Some(Ok(filename))
                 }
             }
-            x => Some(x),
+            Err(x) => Some(Err(x)),
         }
     }
 }
@@ -104,9 +108,6 @@ fn format_time(timestamp: u32, now: i64) -> Result<String> {
     }
 }
 
-// TODO: Document hardlink structure
-type SeenFiles = HashMap<u128, String>;
-
 fn read_cpio_and_print_filenames<R: Read + SeekForward, W: Write>(
     archive: &mut R,
     out: &mut W,
@@ -236,12 +237,13 @@ fn seek_to_cpio_end(archive: &mut File) -> Result<()> {
     Ok(())
 }
 
+/// Return the number of concatenated cpio archives.
 pub fn get_cpio_archive_count(archive: &mut File) -> Result<u32> {
     let mut count = 0;
     loop {
-        let compression = match read_magic_header(archive) {
+        let compression = match read_magic_header(archive)? {
             None => return Ok(count),
-            Some(x) => x?,
+            Some(x) => x,
         };
         count += 1;
         if compression.is_uncompressed() {
@@ -272,58 +274,36 @@ fn get_source_date_epoch() -> Option<u32> {
     }
 }
 
-pub fn print_cpio_archive_count<W: Write>(mut archive: File, out: &mut W) -> Result<()> {
-    let count = get_cpio_archive_count(&mut archive)?;
-    writeln!(out, "{count}")?;
-    Ok(())
-}
-
-// Return the size in bytes of the uncompressed data.
-pub fn create_cpio_archive(
+/// Create a cpio archive and return the size in bytes of the uncompressed data.
+///
+/// **Warning**: This function was designed for the `3cpio` command-line application.
+/// The API can change between releases and no stability promises are given.
+/// Please get in contact to support your use case and make the API for this function stable.
+pub fn create_cpio_archive<W: Write>(
     archive: Option<File>,
-    alignment: Option<u32>,
-    log_level: u32,
+    alignment: Option<NonZeroU32>,
+    logger: &mut Logger<W>,
 ) -> Result<u64> {
     let source_date_epoch = get_source_date_epoch();
     let stdin = std::io::stdin();
     let buf_reader = std::io::BufReader::new(stdin);
-    if log_level >= LOG_LEVEL_DEBUG {
-        eprintln!("Parsing manifest from stdin...");
-    }
-    let manifest = Manifest::from_input(buf_reader, log_level)?;
-    if log_level >= LOG_LEVEL_DEBUG {
-        eprintln!("Writing cpio...");
-    }
-    manifest.write_archive(archive, alignment, source_date_epoch, log_level)
-}
-
-pub fn examine_cpio_content<W: Write>(mut archive: File, out: &mut W) -> Result<()> {
-    loop {
-        let compression = match read_magic_header(&mut archive) {
-            None => return Ok(()),
-            Some(x) => x?,
-        };
-        writeln!(
-            out,
-            "{}\t{}",
-            archive.stream_position()?,
-            compression.command()
-        )?;
-        if compression.is_uncompressed() {
-            seek_to_cpio_end(&mut archive)?;
-        } else {
-            break;
-        }
-    }
-    Ok(())
+    debug!(logger, "Parsing manifest from stdin...")?;
+    let manifest = Manifest::from_input(buf_reader, logger)?;
+    debug!(logger, "Writing cpio...")?;
+    manifest.write_archive(archive, alignment, source_date_epoch, logger)
 }
 
+/// List the contents of the cpio archives.
+///
+/// **Warning**: This function was designed for the `3cpio` command-line application.
+/// The API can change between releases and no stability promises are given.
+/// Please get in contact to support your use case and make the API for this function stable.
 pub fn list_cpio_content<W: Write>(
     mut archive: File,
     out: &mut W,
     parts: Option<&Ranges>,
     patterns: &Vec<Pattern>,
-    log_level: u32,
+    log_level: Level,
 ) -> Result<()> {
     let mut user_group_cache = UserGroupCache::new();
     let now = SystemTime::now()
@@ -335,9 +315,9 @@ pub fn list_cpio_content<W: Write>(
     let mut count = 0;
     loop {
         count += 1;
-        let compression = match read_magic_header(&mut archive) {
+        let compression = match read_magic_header(&mut archive)? {
             None => return Ok(()),
-            Some(x) => x?,
+            Some(x) => x,
         };
         if parts.is_some_and(|f| !f.contains(&count)) {
             if compression.is_uncompressed() && parts.unwrap().has_more(&count) {
@@ -347,28 +327,28 @@ pub fn list_cpio_content<W: Write>(
             break;
         }
         if compression.is_uncompressed() {
-            if log_level >= LOG_LEVEL_INFO {
+            if log_level >= Level::Info {
                 read_cpio_and_print_long_format(
                     &mut archive,
                     out,
                     patterns,
                     now,
                     &mut user_group_cache,
-                    log_level >= LOG_LEVEL_DEBUG,
+                    log_level >= Level::Debug,
                 )?;
             } else {
                 read_cpio_and_print_filenames(&mut archive, out, patterns)?;
             }
         } else {
             let mut decompressed = compression.decompress(archive)?;
-            if log_level >= LOG_LEVEL_INFO {
+            if log_level >= Level::Info {
                 read_cpio_and_print_long_format(
                     &mut decompressed,
                     out,
                     patterns,
                     now,
                     &mut user_group_cache,
-                    log_level >= LOG_LEVEL_DEBUG,
+                    log_level >= Level::Debug,
                 )?;
             } else {
                 read_cpio_and_print_filenames(&mut decompressed, out, patterns)?;
@@ -382,15 +362,15 @@ pub fn list_cpio_content<W: Write>(
 #[cfg(test)]
 mod tests {
     use std::env;
-    use std::io::SeekFrom;
     use std::path::{Path, PathBuf};
 
     use super::*;
+    use crate::logger::Level;
 
     // Lock for tests that rely on / change the current directory
-    pub static TEST_LOCK: std::sync::Mutex<u32> = std::sync::Mutex::new(0);
+    pub(crate) static TEST_LOCK: std::sync::Mutex<u32> = std::sync::Mutex::new(0);
 
-    pub fn tests_path<P: AsRef<Path>>(path: P) -> PathBuf {
+    pub(crate) fn tests_path<P: AsRef<Path>>(path: P) -> PathBuf {
         Path::new(env!("CARGO_MANIFEST_DIR"))
             .join("tests")
             .join(path)
@@ -408,6 +388,13 @@ mod tests {
         }
     }
 
+    #[test]
+    fn test_print_cpio_archive_count_compressed() {
+        let mut archive = File::open(tests_path("zstd.cpio")).expect("test cpio should be present");
+        let count = get_cpio_archive_count(&mut archive).unwrap();
+        assert_eq!(count, 2);
+    }
+
     #[test]
     fn test_get_cpio_archive_count_single() {
         let mut archive =
@@ -425,7 +412,7 @@ mod tests {
             &mut output,
             Some(&"2-".parse::<Ranges>().unwrap()),
             &Vec::new(),
-            LOG_LEVEL_WARNING,
+            Level::Warning,
         )
         .unwrap();
         assert_eq!(
@@ -439,7 +426,7 @@ mod tests {
         let archive = File::open(tests_path("xz.cpio")).unwrap();
         let patterns = vec![Pattern::new("p?th").unwrap()];
         let mut output = Vec::new();
-        list_cpio_content(archive, &mut output, None, &patterns, LOG_LEVEL_WARNING).unwrap();
+        list_cpio_content(archive, &mut output, None, &patterns, Level::Warning).unwrap();
         assert_eq!(String::from_utf8(output).unwrap(), "path\n");
     }
 
@@ -448,23 +435,10 @@ mod tests {
         let archive = File::open(tests_path("single.cpio")).unwrap();
         let patterns = vec![Pattern::new("*/file").unwrap()];
         let mut output = Vec::new();
-        list_cpio_content(archive, &mut output, None, &patterns, LOG_LEVEL_WARNING).unwrap();
+        list_cpio_content(archive, &mut output, None, &patterns, Level::Warning).unwrap();
         assert_eq!(String::from_utf8(output).unwrap(), "path/file\n");
     }
 
-    #[test]
-    fn test_print_cpio_archive_count() {
-        let mut archive = File::open(tests_path("zstd.cpio")).expect("test cpio should be present");
-        let mut output = Vec::new();
-
-        let count = get_cpio_archive_count(&mut archive).unwrap();
-        assert_eq!(count, 2);
-
-        archive.seek(SeekFrom::Start(0)).unwrap();
-        print_cpio_archive_count(archive, &mut output).unwrap();
-        assert_eq!(String::from_utf8(output).unwrap(), "2\n");
-    }
-
     #[test]
     fn test_read_cpio_and_print_long_format_character_device() {
         // Wrapped before mtime and filename
