@@ -3,36 +3,37 @@
 
 use std::env::set_current_dir;
 use std::fs::{create_dir, read_dir, File};
-use std::io::ErrorKind;
+use std::io::{ErrorKind, Write};
+use std::num::NonZeroU32;
 use std::path::Path;
 use std::process::ExitCode;
 
 use glob::Pattern;
 use lexopt::prelude::*;
 
+use threecpio::examine::examine_cpio_content;
 use threecpio::extract::{extract_cpio_archive, ExtractOptions};
+use threecpio::logger::{Level, Logger};
 use threecpio::ranges::Ranges;
-use threecpio::{
-    create_cpio_archive, examine_cpio_content, list_cpio_content, print_cpio_archive_count,
-    LOG_LEVEL_DEBUG, LOG_LEVEL_INFO, LOG_LEVEL_WARNING,
-};
+use threecpio::{create_cpio_archive, get_cpio_archive_count, list_cpio_content};
 
 #[derive(Debug)]
 struct Args {
     count: bool,
     create: bool,
-    data_alignment: Option<u32>,
+    data_alignment: Option<NonZeroU32>,
     directory: String,
     examine: bool,
     extract: bool,
     force: bool,
     list: bool,
-    log_level: u32,
+    log_level: Level,
     archive: Option<String>,
     make_directories: bool,
     parts: Option<Ranges>,
     patterns: Vec<Pattern>,
     preserve_permissions: bool,
+    raw: bool,
     subdir: Option<String>,
     to_stdout: bool,
 }
@@ -55,7 +56,7 @@ fn print_help() {
         "Usage:
     {executable} --count ARCHIVE
     {executable} {{-c|--create}} [-v|--debug] [-C DIR] [--data-align BYTES] [ARCHIVE] < manifest
-    {executable} {{-e|--examine}} ARCHIVE
+    {executable} {{-e|--examine}} [--raw] ARCHIVE
     {executable} {{-t|--list}} [-v|--debug] [-P LIST] ARCHIVE [pattern...]
     {executable} {{-x|--extract}} [-v|--debug] [-C DIR] [--make-directories] [-P LIST] [-p] [-s NAME] [--to-stdout] [--force] ARCHIVE [pattern...]
 
@@ -72,6 +73,7 @@ Optional arguments:
   -p, --preserve-permissions
                  Set permissions of extracted files to those recorded in the
                  archive (default for superuser).
+  --raw          Use a machine-readable output format.
   -s, --subdir   Extract the cpio archives into separate directories (using the
                  given name plus an incrementing number)
   --to-stdout    Extract files to standard output
@@ -99,11 +101,12 @@ fn parse_args() -> Result<Args, lexopt::Error> {
     let mut parts = None;
     let mut preserve_permissions = is_root();
     let mut list = 0;
-    let mut log_level = LOG_LEVEL_WARNING;
+    let mut log_level = Level::Warning;
     let mut directory = ".".into();
     let mut archive = None;
     let mut make_directories = false;
     let mut patterns = Vec::new();
+    let mut raw = false;
     let mut subdir: Option<String> = None;
     let mut to_stdout = false;
     let mut arguments = Vec::new();
@@ -121,8 +124,8 @@ fn parse_args() -> Result<Args, lexopt::Error> {
             }
             Long("data-align") => {
                 let value = parser.value()?;
-                data_alignment = if let Ok(int_value) = value.parse() {
-                    if int_value % 4 != 0 || int_value == 0 {
+                data_alignment = if let Ok(int_value) = value.parse::<NonZeroU32>() {
+                    if int_value.get() % 4 != 0 {
                         return Err("--data-align must be a multiple of 4 bytes".into());
                     }
                     Some(int_value)
@@ -131,7 +134,7 @@ fn parse_args() -> Result<Args, lexopt::Error> {
                 };
             }
             Long("debug") => {
-                log_level = LOG_LEVEL_DEBUG;
+                log_level = Level::Debug;
             }
             Short('e') | Long("examine") => {
                 examine = 1;
@@ -152,6 +155,9 @@ fn parse_args() -> Result<Args, lexopt::Error> {
             Short('p') | Long("preserve-permissions") => {
                 preserve_permissions = true;
             }
+            Long("raw") => {
+                raw = true;
+            }
             Short('s') | Long("subdir") => {
                 subdir = Some(parser.value()?.string()?);
             }
@@ -162,8 +168,8 @@ fn parse_args() -> Result<Args, lexopt::Error> {
                 to_stdout = true;
             }
             Short('v') | Long("verbose") => {
-                if log_level <= LOG_LEVEL_INFO {
-                    log_level = LOG_LEVEL_INFO;
+                if log_level <= Level::Info {
+                    log_level = Level::Info;
                 }
             }
             Short('V') | Long("version") => {
@@ -223,6 +229,7 @@ fn parse_args() -> Result<Args, lexopt::Error> {
         parts,
         patterns,
         preserve_permissions,
+        raw,
         subdir,
         to_stdout,
     })
@@ -268,6 +275,13 @@ fn create_and_set_current_dir(path: &str, force: bool) -> Result<(), String> {
     Ok(())
 }
 
+/// Print the number of concatenated cpio archives.
+fn print_cpio_archive_count<W: Write>(mut archive: File, out: &mut W) -> std::io::Result<()> {
+    let count = get_cpio_archive_count(&mut archive)?;
+    writeln!(out, "{count}")?;
+    Ok(())
+}
+
 fn main() -> ExitCode {
     let executable = std::env::args().next().unwrap();
     let args = match parse_args() {
@@ -277,6 +291,7 @@ fn main() -> ExitCode {
             return ExitCode::from(2);
         }
     };
+    let mut logger = Logger::new_stderr(args.log_level.clone());
 
     if args.create {
         let mut archive = None;
@@ -288,7 +303,7 @@ fn main() -> ExitCode {
                     return ExitCode::FAILURE;
                 }
             };
-            if args.log_level >= LOG_LEVEL_DEBUG {
+            if args.log_level >= Level::Debug {
                 eprintln!("{executable}: Opened '{path}' for writing.");
             }
         }
@@ -299,7 +314,7 @@ fn main() -> ExitCode {
             );
             return ExitCode::FAILURE;
         }
-        let result = create_cpio_archive(archive, args.data_alignment, args.log_level);
+        let result = create_cpio_archive(archive, args.data_alignment, &mut logger);
         if let Err(error) = result {
             match error.kind() {
                 ErrorKind::BrokenPipe => {}
@@ -342,7 +357,7 @@ fn main() -> ExitCode {
     } else if args.examine {
         (
             "examine content",
-            examine_cpio_content(archive, &mut stdout),
+            examine_cpio_content(archive, &mut stdout, args.raw),
         )
     } else if args.extract {
         (
@@ -351,7 +366,7 @@ fn main() -> ExitCode {
                 archive,
                 args.to_stdout.then_some(&mut stdout),
                 &args.extract_options(),
-                args.log_level,
+                &mut logger,
             ),
         )
     } else if args.list {
